@@ -34,6 +34,11 @@
 #define O_BINARY 0
 #endif
 
+/* This property changed names in openchange, try to support both */
+#ifndef PidLidTaskAcceptanceState
+	#define PidLidTaskAcceptanceState PidLidAcceptanceState
+#endif
+
 #define d(x) 
 
 static gboolean appt_build_name_id (ExchangeMapiConnection *conn, mapi_id_t fid, TALLOC_CTX *mem_ctx, struct SPropTagArray *props);
@@ -368,9 +373,14 @@ exchange_mapi_cal_util_fetch_recipients (ECalComponent *comp, GSList **recip_lis
 		const gchar *str = NULL;
 		icalparameter *param;
 
+		str = icalproperty_get_attendee (att_prop);
+		if (!str || g_ascii_strcasecmp (str, org) == 0) {
+			att_prop = icalcomponent_get_next_property (icalcomp, ICAL_ATTENDEE_PROPERTY);
+			continue;
+		}
+
 		recipient = g_new0 (ExchangeMAPIRecipient, 1);
 
-		str = icalproperty_get_attendee (att_prop);
 		if (!g_ascii_strncasecmp (str, "mailto:", 7))
 			recipient->email_id = (str) + 7;
 		else
@@ -680,14 +690,15 @@ id_to_string (GByteArray *ba)
 }
 
 ECalComponent *
-exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcomponent_kind kind, const gchar *mid, struct mapi_SPropValue_array *properties,
+exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, mapi_id_t fid, icalcomponent_kind kind, const gchar *mid, struct mapi_SPropValue_array *properties,
 					   GSList *streams, GSList *recipients, GSList *attachments,
-					   const gchar *local_store_uri, const icaltimezone *default_zone)
+					   const gchar *local_store_uri, const icaltimezone *default_zone, gboolean is_reply, GSList **detached_components)
 {
 	ECalComponent *comp = NULL;
 	struct timeval t;
 	ExchangeMAPIStream *body_stream;
 	const gchar *subject = NULL, *body = NULL;
+	const struct StringArray_r *categories_array;
 	const uint32_t *ui32;
 	const bool *b;
 	icalcomponent *ical_comp;
@@ -742,13 +753,33 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 	icalcomponent_set_summary (ical_comp, subject);
 	icalcomponent_set_description (ical_comp, body);
 
+	categories_array = exchange_mapi_util_find_array_namedid (properties, conn, fid, PidNameKeywords);
+	if (categories_array) {
+		GSList *categories = NULL;
+		gint ii;
+
+		for (ii = 0; ii < categories_array->cValues; ii++) {
+			const gchar *category = categories_array->lppszA[ii];
+
+			if (!category || !*category)
+				continue;
+
+			categories = g_slist_append (categories, (gpointer) category);
+		}
+
+		e_cal_component_set_categories_list (comp, categories);
+
+		g_slist_free (categories);
+	}
+
 	if (icalcomponent_isa (ical_comp) == ICAL_VEVENT_COMPONENT) {
 		const gchar *location = NULL;
 		const gchar *dtstart_tz_location = NULL, *dtend_tz_location = NULL;
+		gboolean all_day;
 		ExchangeMAPIStream *stream;
 
 		/* CleanGlobalObjectId */
-		stream = exchange_mapi_util_find_stream (streams, PROP_TAG(PT_BINARY, 0x0023));
+		stream = exchange_mapi_util_find_stream (streams, PidLidCleanGlobalObjectId);
 		if (stream) {
 			gchar *value = id_to_string (stream->value);
 			prop = icalproperty_new_x (value);
@@ -758,17 +789,27 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 		}
 
 		/* GlobalObjectId */
-		stream = exchange_mapi_util_find_stream (streams, PROP_TAG(PT_BINARY, 0x0003));
+		stream = exchange_mapi_util_find_stream (streams, PidLidGlobalObjectId);
 		if (stream) {
 			gchar *value = id_to_string (stream->value);
 			prop = icalproperty_new_x (value);
 			icalproperty_set_x_name (prop, "X-EVOLUTION-MAPI-GLOBALID");
 			icalcomponent_add_property (ical_comp, prop);
+			if (value && *value) {
+				e_cal_component_set_uid (comp, value);
+
+				if (!g_str_equal (value, mid)) {
+					prop = icalproperty_new_x (mid);
+					icalproperty_set_x_name (prop, "X-EVOLUTION-MAPI-MID");
+					icalcomponent_add_property (ical_comp, prop);
+				}
+			}
+
 			g_free (value);
 		}
 
 		/* AppointmentSequence */
-		ui32 = (const uint32_t *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_LONG, 0x8201));
+		ui32 = (const uint32_t *)find_mapi_SPropValue_data(properties, PidLidAppointmentSequence);
 		if (ui32) {
 			gchar *value = g_strdup_printf ("%d", *ui32);
 			prop = icalproperty_new_x (value);
@@ -777,47 +818,48 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 			g_free (value);
 		}
 
-		location = (const gchar *)exchange_mapi_util_find_array_propval(properties, PROP_TAG(PT_STRING8, 0x8208));
+		location = (const gchar *)exchange_mapi_util_find_array_propval(properties, PidLidLocation);
 		if (location && *location)
 			icalcomponent_set_location (ical_comp, location);
 
-		b = (const bool *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_BOOLEAN, 0x8215));
+		b = (const bool *)find_mapi_SPropValue_data(properties, PidLidAppointmentSubType);
+		all_day = b && *b;
 
-		stream = exchange_mapi_util_find_stream (streams, PROP_TAG(PT_BINARY, 0x825E));
+		stream = exchange_mapi_util_find_stream (streams, PidLidAppointmentTimeZoneDefinitionStartDisplay);
 		if (stream) {
 			gchar *buf = exchange_mapi_cal_util_bin_to_mapi_tz (stream->value);
 			dtstart_tz_location = exchange_mapi_cal_tz_util_get_ical_equivalent (buf);
 			g_free (buf);
 		}
 
-		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PROP_TAG(PT_SYSTIME, 0x820D)) == MAPI_E_SUCCESS) {
+		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PidLidAppointmentStartWhole) == MAPI_E_SUCCESS) {
 			icaltimezone *zone = dtstart_tz_location ? icaltimezone_get_builtin_timezone (dtstart_tz_location) : (icaltimezone *)default_zone;
-			prop = icalproperty_new_dtstart (icaltime_from_timet_with_zone (t.tv_sec, (b && *b), zone));
-			if (zone && icaltimezone_get_tzid (zone)) {
+			prop = icalproperty_new_dtstart (icaltime_from_timet_with_zone (t.tv_sec, all_day, zone));
+			if (!all_day && zone && icaltimezone_get_tzid (zone)) {
 				icalproperty_add_parameter (prop, icalparameter_new_tzid (icaltimezone_get_tzid (zone)));
 			}
 
 			icalcomponent_add_property (ical_comp, prop);
 		}
 
-		stream = exchange_mapi_util_find_stream (streams, PROP_TAG(PT_BINARY, 0x825F));
+		stream = exchange_mapi_util_find_stream (streams, PidLidAppointmentTimeZoneDefinitionEndDisplay);
 		if (stream) {
 			gchar *buf = exchange_mapi_cal_util_bin_to_mapi_tz (stream->value);
 			dtend_tz_location = exchange_mapi_cal_tz_util_get_ical_equivalent (buf);
 			g_free (buf);
 		}
 
-		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PROP_TAG(PT_SYSTIME, 0x820E)) == MAPI_E_SUCCESS) {
+		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PidLidAppointmentEndWhole) == MAPI_E_SUCCESS) {
 			icaltimezone *zone = dtend_tz_location ? icaltimezone_get_builtin_timezone (dtend_tz_location) : (icaltimezone *)default_zone;
-			prop = icalproperty_new_dtend (icaltime_from_timet_with_zone (t.tv_sec, (b && *b), zone));
-			if (zone && icaltimezone_get_tzid (zone)) {
+			prop = icalproperty_new_dtend (icaltime_from_timet_with_zone (t.tv_sec, all_day, zone));
+			if (!all_day && zone && icaltimezone_get_tzid (zone)) {
 				icalproperty_add_parameter (prop, icalparameter_new_tzid (icaltimezone_get_tzid (zone)));
 			}
 
 			icalcomponent_add_property (ical_comp, prop);
 		}
 
-		ui32 = (const uint32_t *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_LONG, 0x8205));
+		ui32 = (const uint32_t *)find_mapi_SPropValue_data(properties, PidLidBusyStatus);
 		if (ui32) {
 			prop = icalproperty_new_transp (get_transp_from_prop (*ui32));
 			icalcomponent_add_property (ical_comp, prop);
@@ -826,9 +868,75 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 		if (recipients) {
 			b = (const bool *)find_mapi_SPropValue_data(properties, PR_RESPONSE_REQUESTED);
 			ical_attendees_from_props (ical_comp, recipients, (b && *b));
-			if (icalcomponent_get_first_property (ical_comp, ICAL_ORGANIZER_PROPERTY) == NULL) {
+			if (is_reply) {
+				if (icalcomponent_get_first_property (ical_comp, ICAL_ORGANIZER_PROPERTY) == NULL) {
+					gchar *val, *to_free = NULL;
+					const gchar *name = exchange_mapi_util_find_array_propval (properties, PR_RCVD_REPRESENTING_NAME_UNICODE);
+					const gchar *email_type = exchange_mapi_util_find_array_propval (properties, PR_RCVD_REPRESENTING_ADDRTYPE_UNICODE);
+					const gchar *email = exchange_mapi_util_find_array_propval (properties, PR_RCVD_REPRESENTING_EMAIL_ADDRESS_UNICODE);
+
+					if (!name)
+						name = "";
+					if (!email_type)
+						email_type = "";
+					if (!email)
+						email = "";
+
+					if (g_str_equal (email_type, "EX")) {
+						to_free = exchange_mapi_connection_ex_to_smtp (conn, email, NULL, NULL);
+						email = to_free;
+					}
+
+					val = g_strdup_printf ("MAILTO:%s", email);
+					prop = icalproperty_new_organizer (val);
+					g_free (val);
+
+					/* CN */
+					param = icalparameter_new_cn (name);
+					icalproperty_add_parameter (prop, param);
+
+					icalcomponent_add_property (ical_comp, prop);
+
+					g_free (to_free);
+				}
+
+				if (icalcomponent_get_first_property (ical_comp, ICAL_ATTENDEE_PROPERTY) == NULL) {
+					const uint32_t *ui32;
+					gchar *val, *to_free = NULL;
+					const gchar *name = exchange_mapi_util_find_array_propval (properties, PR_SENT_REPRESENTING_NAME_UNICODE);
+					const gchar *email_type = exchange_mapi_util_find_array_propval (properties, PR_SENT_REPRESENTING_ADDRTYPE_UNICODE);
+					const gchar *email = exchange_mapi_util_find_array_propval (properties, PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE);
+
+					if (!name)
+						name = "";
+					if (!email_type)
+						email_type = "";
+					if (!email)
+						email = "";
+
+					if (g_str_equal (email_type, "EX")) {
+						to_free = exchange_mapi_connection_ex_to_smtp (conn, email, NULL, NULL);
+						email = to_free;
+					}
+
+					val = g_strdup_printf ("MAILTO:%s", email);
+					prop = icalproperty_new_attendee (val);
+					g_free (val);
+
+					/* CN */
+					param = icalparameter_new_cn (name);
+					icalproperty_add_parameter (prop, param);
+
+					ui32 = exchange_mapi_util_find_array_propval (properties, PidLidResponseStatus);
+					param = icalparameter_new_partstat (get_partstat_from_trackstatus (ui32 ? *ui32 : olResponseNone));
+					icalproperty_add_parameter (prop, param);
+
+					icalcomponent_add_property (ical_comp, prop);
+
+					g_free (to_free);
+				}
+			} else if (icalcomponent_get_first_property (ical_comp, ICAL_ORGANIZER_PROPERTY) == NULL) {
 				gchar *val, *sender_free = NULL, *sent_free = NULL;
-//				const gchar *sender_name = (const gchar *) exchange_mapi_util_find_array_propval (properties, PR_SENDER_NAME_UNICODE);
 				const gchar *sender_email_type = (const gchar *) exchange_mapi_util_find_array_propval (properties, PR_SENDER_ADDRTYPE_UNICODE);
 				const gchar *sender_email = (const gchar *) exchange_mapi_util_find_array_propval (properties, PR_SENDER_EMAIL_ADDRESS_UNICODE);
 				const gchar *sent_name = (const gchar *) exchange_mapi_util_find_array_propval (properties, PR_SENT_REPRESENTING_NAME_UNICODE);
@@ -836,11 +944,11 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 				const gchar *sent_email = (const gchar *) exchange_mapi_util_find_array_propval (properties, PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE);
 
 				if (!g_utf8_collate (sender_email_type, "EX")) {
-					sender_free = exchange_mapi_connection_ex_to_smtp (conn, sender_email, NULL);
+					sender_free = exchange_mapi_connection_ex_to_smtp (conn, sender_email, NULL, NULL);
 					sender_email = sender_free;
 				}
 				if (!g_utf8_collate (sent_email_type, "EX")) {
-					sent_free = exchange_mapi_connection_ex_to_smtp (conn, sent_email, NULL);
+					sent_free = exchange_mapi_connection_ex_to_smtp (conn, sent_email, NULL, NULL);
 					sent_email = sent_free;
 				}
 
@@ -865,20 +973,20 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 			}
 		}
 
-		b = (const bool *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_BOOLEAN, 0x8223));
+		b = (const bool *)find_mapi_SPropValue_data(properties, PidLidRecurring);
 		if (b && *b) {
-			stream = exchange_mapi_util_find_stream (streams, PROP_TAG(PT_BINARY, 0x8216));
+			stream = exchange_mapi_util_find_stream (streams, PidLidAppointmentRecur);
 			if (stream) {
-				exchange_mapi_cal_util_bin_to_rrule (stream->value, comp);
+				exchange_mapi_cal_util_bin_to_rrule (stream->value, comp, detached_components);
 			}
 		}
 
-		b = (const bool *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_BOOLEAN, 0x8503));
+		b = (const bool *)find_mapi_SPropValue_data(properties, PidLidReminderSet);
 		if (b && *b) {
 			struct timeval start, displaytime;
 
-			if ((get_mapi_SPropValue_array_date_timeval (&start, properties, PROP_TAG(PT_SYSTIME, 0x8502)) == MAPI_E_SUCCESS)
-			 && (get_mapi_SPropValue_array_date_timeval (&displaytime, properties, PROP_TAG(PT_SYSTIME, 0x8560)) == MAPI_E_SUCCESS)) {
+			if ((get_mapi_SPropValue_array_date_timeval (&start, properties, PidLidReminderTime) == MAPI_E_SUCCESS)
+			 && (get_mapi_SPropValue_array_date_timeval (&displaytime, properties, PidLidReminderSignalTime) == MAPI_E_SUCCESS)) {
 				ECalComponentAlarm *e_alarm = e_cal_component_alarm_new ();
 				ECalComponentAlarmTrigger trigger;
 
@@ -899,38 +1007,38 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 		const uint64_t *status = NULL;
 
 		/* NOTE: Exchange tasks are DATE values, not DATE-TIME values, but maybe someday, we could expect Exchange to support it;) */
-		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PROP_TAG(PT_SYSTIME, 0x8104)) == MAPI_E_SUCCESS)
+		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PidLidTaskStartDate) == MAPI_E_SUCCESS)
 			icalcomponent_set_dtstart (ical_comp, icaltime_from_timet_with_zone (t.tv_sec, 1, default_zone));
-		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PROP_TAG(PT_SYSTIME, 0x8105)) == MAPI_E_SUCCESS)
+		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PidLidTaskDueDate) == MAPI_E_SUCCESS)
 			icalcomponent_set_due (ical_comp, icaltime_from_timet_with_zone (t.tv_sec, 1, default_zone));
 
-		status = (const uint64_t *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_LONG, 0x8101));
+		status = (const uint64_t *)find_mapi_SPropValue_data(properties, PidLidTaskStatus);
 		if (status) {
 			icalcomponent_set_status (ical_comp, get_taskstatus_from_prop(*status));
 			if (*status == olTaskComplete
-			&& get_mapi_SPropValue_array_date_timeval (&t, properties, PROP_TAG(PT_SYSTIME, 0x810F)) == MAPI_E_SUCCESS) {
+			&& get_mapi_SPropValue_array_date_timeval (&t, properties, PidLidTaskDateCompleted) == MAPI_E_SUCCESS) {
 				prop = icalproperty_new_completed (icaltime_from_timet_with_zone (t.tv_sec, 1, default_zone));
 				icalcomponent_add_property (ical_comp, prop);
 			}
 		}
 
-		complete = (const double *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_DOUBLE, 0x8102));
+		complete = (const double *)find_mapi_SPropValue_data(properties, PidLidPercentComplete);
 		if (complete) {
 			prop = icalproperty_new_percentcomplete ((gint)(*complete * 100 + 1e-9));
 			icalcomponent_add_property (ical_comp, prop);
 		}
 
-		b = (const bool *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_BOOLEAN, 0x8126));
+		b = (const bool *)find_mapi_SPropValue_data(properties, PidLidTaskFRecurring);
 		if (b && *b) {
 			/* FIXME: Evolution does not support recurring tasks */
 			g_warning ("Encountered a recurring task.");
 		}
 
-		b = (const bool *)find_mapi_SPropValue_data(properties, PROP_TAG(PT_BOOLEAN, 0x8503));
+		b = (const bool *)find_mapi_SPropValue_data(properties, PidLidReminderSet);
 		if (b && *b) {
 			struct timeval abs;
 
-			if (get_mapi_SPropValue_array_date_timeval (&abs, properties, PROP_TAG(PT_SYSTIME, 0x8502)) == MAPI_E_SUCCESS) {
+			if (get_mapi_SPropValue_array_date_timeval (&abs, properties, PidLidReminderTime) == MAPI_E_SUCCESS) {
 				ECalComponentAlarm *e_alarm = e_cal_component_alarm_new ();
 				ECalComponentAlarmTrigger trigger;
 
@@ -966,8 +1074,6 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 		icalcomponent_add_property (ical_comp, prop);
 	}
 
-	/* FIXME: categories */
-
 	set_attachments_to_cal_component (comp, attachments, local_store_uri);
 
 	e_cal_component_rescan (comp);
@@ -986,6 +1092,7 @@ fetch_camel_cal_comp_cb (FetchItemsCallbackData *item_data, gpointer data)
 {
 	struct fetch_camel_cal_data *fccd = data;
 	ECalComponent *comp = NULL;
+	GSList *detached_recurrences = NULL, *d_i = NULL;
 	mapi_id_t mid = 0;
 	icalcomponent *icalcomp = NULL;
 	gchar *str = NULL, *smid = NULL, *filepath;
@@ -1003,10 +1110,11 @@ fetch_camel_cal_comp_cb (FetchItemsCallbackData *item_data, gpointer data)
 			smid = exchange_mapi_util_mapi_id_to_string (item_data->mid);
 		else
 			smid = e_cal_component_gen_uid();
-		comp = exchange_mapi_cal_util_mapi_props_to_comp (item_data->conn, fccd->kind, smid,
+		comp = exchange_mapi_cal_util_mapi_props_to_comp (item_data->conn, item_data->fid, fccd->kind, smid,
 							item_data->properties, item_data->streams, item_data->recipients,
-							item_data->attachments, filepath, NULL);
-		e_cal_component_set_uid (comp, smid);
+							item_data->attachments, filepath, NULL, fccd->method == ICAL_METHOD_REPLY,
+							&detached_recurrences);
+
 		g_free (smid);
 	}
 
@@ -1017,10 +1125,16 @@ fetch_camel_cal_comp_cb (FetchItemsCallbackData *item_data, gpointer data)
 	if (comp)
 		icalcomponent_add_component (icalcomp,
 			icalcomponent_new_clone(e_cal_component_get_icalcomponent(comp)));
+	for (d_i = detached_recurrences; d_i; d_i = g_slist_next (d_i)) {
+		icalcomponent_add_component (icalcomp,
+				icalcomponent_new_clone (e_cal_component_get_icalcomponent (d_i->data)));
+		g_object_unref (d_i->data);
+	}
 	str = icalcomponent_as_ical_string_r (icalcomp);
 	icalcomponent_free (icalcomp);
 	if (comp)
 		g_object_unref (comp);
+	g_slist_free (detached_recurrences);
 
 	exchange_mapi_util_free_stream_list (&item_data->streams);
 	exchange_mapi_util_free_recipient_list (&item_data->recipients);
@@ -1083,7 +1197,10 @@ exchange_mapi_cal_utils_add_named_ids (ExchangeMapiConnection *conn, mapi_id_t f
 		{ PidLidCommonStart, 0 },
 		{ PidLidCommonEnd, 0 },
 		{ PidLidTaskMode, 0 },
-		{ PidLidReminderSignalTime, 0 }
+		{ PidLidReminderSignalTime, 0 },
+		{ PidLidTimeZoneStruct, 0 },
+		{ PidLidTimeZoneDescription, 0 },
+		{ PidNameKeywords, 0 }
 	};
 	icalcomponent_kind kind = pkind;
 
@@ -1167,7 +1284,7 @@ task_build_name_id (ExchangeMapiConnection *conn, mapi_id_t fid, TALLOC_CTX *mem
 		{ PidLidTaskFRecurring, 0 },
 		{ PidLidTaskRole, 0 },
 		{ PidLidTaskOwnership, 0 },
-		{ PidLidAcceptanceState, 0 }
+		{ PidLidTaskAcceptanceState, 0 }
 	};
 
 	if (!props)
@@ -1235,10 +1352,12 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 	uint32_t flag32;
 	bool b;
 	icalproperty *prop;
-	struct icaltimetype dtstart, dtend, utc_dtstart, utc_dtend;
+	struct icaltimetype dtstart, dtend, utc_dtstart, utc_dtend, all_day_dtstart = {0}, all_day_dtend = {0};
 	const icaltimezone *utc_zone;
 	const gchar *dtstart_tz_location, *dtend_tz_location, *text = NULL;
-	struct timeval t;
+	time_t tt;
+	gboolean is_all_day;
+	GSList *categories = NULL;
 
 	g_return_val_if_fail (conn != NULL, FALSE);
 	g_return_val_if_fail (mem_ctx != NULL, FALSE);
@@ -1274,12 +1393,8 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 
 	#define set_datetime_value(hex, dtval) G_STMT_START {		\
 		struct FILETIME	filetime;				\
-		NTTIME		nttime;					\
 									\
-		nttime = timeval_to_nttime(dtval);				\
-									\
-		filetime.dwLowDateTime = (nttime << 32) >> 32;		\
-		filetime.dwHighDateTime = nttime >> 32;			\
+		exchange_mapi_util_time_t_to_filetime (dtval, &filetime); \
 									\
 		if (!exchange_mapi_utils_add_spropvalue (mem_ctx, values, n_values, hex, &filetime)) \
 			return FALSE;	\
@@ -1287,12 +1402,8 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 
 	#define set_named_datetime_value(named_id, dtval) G_STMT_START { \
 		struct FILETIME	filetime;				\
-		NTTIME		nttime;					\
 									\
-		nttime = timeval_to_nttime(dtval);				\
-									\
-		filetime.dwLowDateTime = (nttime << 32) >> 32;		\
-		filetime.dwHighDateTime = nttime >> 32;			\
+		exchange_mapi_util_time_t_to_filetime (dtval, &filetime); \
 									\
 		if (!exchange_mapi_utils_add_spropvalue_named_id (conn, fid, mem_ctx, values, n_values, named_id, &filetime)) \
 			return FALSE;	\
@@ -1314,6 +1425,38 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 	dtstart_tz_location = get_tzid_location (icaltime_get_tzid (dtstart), cbdata);
 	dtend_tz_location = get_tzid_location (icaltime_get_tzid (dtend), cbdata);
 
+	is_all_day = kind == ICAL_VEVENT_COMPONENT && icaltime_is_date (dtstart) && icaltime_is_date (dtend);
+	if (is_all_day) {
+		const gchar *def_location;
+		icaltimezone *use_zone = NULL;
+
+		/* all-day events expect times not in UTC but in local time;
+		   if this differs from the server timezone, then the event
+		   is shown spread among (two) days */
+		def_location = get_tzid_location ("*default-zone*", cbdata);
+		if (def_location && *def_location)
+			use_zone = icaltimezone_get_builtin_timezone (def_location);
+
+		if (!use_zone)
+			use_zone = (icaltimezone *) utc_zone;
+
+		dtstart.is_date = 0;
+		dtstart.hour = 0;
+		dtstart.minute = 0;
+		dtstart.second = 0;
+		all_day_dtstart = icaltime_convert_to_zone (dtstart, use_zone);
+		dtstart.is_date = 1;
+		all_day_dtstart = icaltime_convert_to_zone (all_day_dtstart, (icaltimezone *) utc_zone);
+
+		dtend.is_date = 0;
+		dtend.hour = 0;
+		dtend.minute = 0;
+		dtend.second = 0;
+		all_day_dtend = icaltime_convert_to_zone (dtend, use_zone);
+		dtend.is_date = 1;
+		all_day_dtend = icaltime_convert_to_zone (all_day_dtend, (icaltimezone *) utc_zone);
+	}
+
 	utc_dtstart = icaltime_convert_to_zone (dtstart, (icaltimezone *)utc_zone);
 	utc_dtend = icaltime_convert_to_zone (dtend, (icaltimezone *)utc_zone);
 
@@ -1337,6 +1480,33 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 	set_value (PR_BODY_UNICODE, text);
 	text = NULL;
 
+	e_cal_component_get_categories_list (comp, &categories);
+	if (categories) {
+		gint ii;
+		GSList *c;
+		struct StringArray_r *categories_array;
+
+		categories_array = talloc_zero (mem_ctx, struct StringArray_r);
+		categories_array->cValues = g_slist_length (categories);
+		categories_array->lppszA = (const char **) talloc_zero_array (mem_ctx, gchar *, categories_array->cValues);
+
+		for (c = categories, ii = 0; c; c = c->next, ii++) {
+			const gchar *category = c->data;
+
+			if (!category || !*category) {
+				ii--;
+				categories_array->cValues--;
+				continue;
+			}
+
+			categories_array->lppszA[ii] = talloc_strdup (mem_ctx, category);
+		}
+
+		set_named_value (PidNameKeywords, categories_array);
+
+		e_cal_component_free_categories_list (categories);
+	}
+
 	/* Priority and Importance */
 	prop = icalcomponent_get_first_property (ical_comp, ICAL_PRIORITY_PROPERTY);
 	flag32 = prop ? get_prio_prop_from_priority (icalproperty_get_priority (prop)) : PRIORITY_NORMAL;
@@ -1344,12 +1514,17 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 	flag32 = prop ? get_imp_prop_from_priority (icalproperty_get_priority (prop)) : IMPORTANCE_NORMAL;
 	set_value (PR_IMPORTANCE, &flag32);
 
-	set_value (PR_SENT_REPRESENTING_NAME_UNICODE, cbdata->ownername);
-	set_value (PR_SENT_REPRESENTING_ADDRTYPE_UNICODE, cbdata->owneridtype);
-	set_value (PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE, cbdata->ownerid);
-	set_value (PR_SENDER_NAME_UNICODE, cbdata->username);
-	set_value (PR_SENDER_ADDRTYPE_UNICODE, cbdata->useridtype);
-	set_value (PR_SENDER_EMAIL_ADDRESS_UNICODE, cbdata->userid);
+	if (cbdata->ownername && cbdata->owneridtype && cbdata->ownerid) {
+		set_value (PR_SENT_REPRESENTING_NAME_UNICODE, cbdata->ownername);
+		set_value (PR_SENT_REPRESENTING_ADDRTYPE_UNICODE, cbdata->owneridtype);
+		set_value (PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE, cbdata->ownerid);
+	}
+
+	if (cbdata->username && cbdata->useridtype && cbdata->userid) {
+		set_value (PR_SENDER_NAME_UNICODE, cbdata->username);
+		set_value (PR_SENDER_ADDRTYPE_UNICODE, cbdata->useridtype);
+		set_value (PR_SENDER_EMAIL_ADDRESS_UNICODE, cbdata->userid);
+	}
 
 	flag32 = cbdata->msgflags;
 	set_value (PR_MESSAGE_FLAGS, &flag32);
@@ -1392,13 +1567,11 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 		}
 	set_named_value (PidLidReminderSet, &b);
 	set_named_value (PidLidReminderDelta, &flag32);
-	t.tv_sec = icaltime_as_timet (utc_dtstart);
-	t.tv_usec = 0;
-	set_named_datetime_value (PidLidReminderTime, &t);
-	t.tv_sec = icaltime_as_timet (utc_dtstart) - (flag32 * SECS_IN_MINUTE);
-	t.tv_usec = 0;
+	tt = icaltime_as_timet (utc_dtstart);
+	set_named_datetime_value (PidLidReminderTime, tt);
+	tt = icaltime_as_timet (utc_dtstart) - (flag32 * SECS_IN_MINUTE);
 	/* ReminderNextTime: FIXME for recurrence */
-	set_named_datetime_value (PidLidReminderSignalTime, &t);
+	set_named_datetime_value (PidLidReminderSignalTime, tt);
 
 	/* Sensitivity, Private */
 	flag32 = olNormal;	/* default */
@@ -1411,15 +1584,13 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 	set_value (PR_SENSITIVITY, &flag32);
 	set_named_value (PidLidPrivate, &b);
 
-	t.tv_sec = icaltime_as_timet (utc_dtstart);
-	t.tv_usec = 0;
-	set_named_datetime_value (PidLidCommonStart, &t);
-	set_datetime_value (PR_START_DATE, &t);
+	tt = icaltime_as_timet (is_all_day ? all_day_dtstart : utc_dtstart);
+	set_named_datetime_value (PidLidCommonStart, tt);
+	set_datetime_value (PR_START_DATE, tt);
 
-	t.tv_sec = icaltime_as_timet (utc_dtend);
-	t.tv_usec = 0;
-	set_named_datetime_value (PidLidCommonEnd, &t);
-	set_datetime_value (PR_END_DATE, &t);
+	tt = icaltime_as_timet (is_all_day ? all_day_dtend : utc_dtend);
+	set_named_datetime_value (PidLidCommonEnd, tt);
+	set_datetime_value (PR_END_DATE, tt);
 
 	b = 1;
 	set_value (PR_RESPONSE_REQUESTED, &b);
@@ -1465,12 +1636,15 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 		b = 0;
 		set_named_value (PidLidAutoFillLocation, &b);
 
+		/* All-day event */
+		b = is_all_day ? 1 : 0;
+		set_named_value (PidLidAppointmentSubType, &b);
+
 		/* Start */
-		t.tv_sec = icaltime_as_timet (utc_dtstart);
-		t.tv_usec = 0;
-		set_named_datetime_value (PidLidAppointmentStartWhole, &t);
+		tt = icaltime_as_timet (is_all_day ? all_day_dtstart : utc_dtstart);
+		set_named_datetime_value (PidLidAppointmentStartWhole, tt);
 		/* FIXME: for recurrence */
-		set_named_datetime_value (PidLidClipStart, &t);
+		set_named_datetime_value (PidLidClipStart, tt);
 
 		/* Start TZ */
 		mapi_tzid = exchange_mapi_cal_tz_util_get_mapi_equivalent ((dtstart_tz_location && *dtstart_tz_location) ? dtstart_tz_location : "UTC");
@@ -1478,13 +1652,13 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 			exchange_mapi_cal_util_mapi_tz_to_bin (mapi_tzid, &start_tz);
 			set_named_value (PidLidAppointmentTimeZoneDefinitionStartDisplay, &start_tz);
 		}
+		set_named_value (PidLidTimeZoneDescription, mapi_tzid ? mapi_tzid : "");
 
 		/* End */
-		t.tv_sec = icaltime_as_timet (utc_dtend);
-		t.tv_usec = 0;
-		set_named_datetime_value (PidLidAppointmentEndWhole, &t);
+		tt = icaltime_as_timet (is_all_day ? all_day_dtend : utc_dtend);
+		set_named_datetime_value (PidLidAppointmentEndWhole, tt);
 		/* FIXME: for recurrence */
-		set_named_datetime_value (PidLidClipEnd, &t);
+		set_named_datetime_value (PidLidClipEnd, tt);
 
 		/* End TZ */
 		mapi_tzid = exchange_mapi_cal_tz_util_get_mapi_equivalent ((dtend_tz_location && *dtend_tz_location) ? dtend_tz_location : "UTC");
@@ -1498,10 +1672,6 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 		flag32 /= MINUTES_IN_HOUR;
 		set_named_value (PidLidAppointmentDuration, &flag32);
 
-		/* All-day event */
-		b = (icaltime_is_date (dtstart) && icaltime_is_date (dtend));
-		set_named_value (PidLidAppointmentSubType, &b);
-
 		if (e_cal_component_has_recurrences (comp)) {
 			GSList *rrule_list = NULL;
 			struct icalrecurrencetype *rt = NULL;
@@ -1531,8 +1701,13 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 		flag32 = cbdata->appt_seq;
 		set_named_value (PidLidAppointmentSequence, &flag32);
 
-		set_named_value (PidLidCleanGlobalObjectId, cbdata->cleanglobalid);
-		set_named_value (PidLidGlobalObjectId, cbdata->globalid);
+		if (cbdata->cleanglobalid) {
+			set_named_value (PidLidCleanGlobalObjectId, cbdata->cleanglobalid);
+		}
+
+		if (cbdata->globalid) {
+			set_named_value (PidLidGlobalObjectId, cbdata->globalid);
+		}
 
 		flag32 = cbdata->resp;
 		set_named_value (PidLidResponseStatus, &flag32);
@@ -1725,25 +1900,22 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 			completed = icalproperty_get_completed (prop);
 
 			completed.hour = completed.minute = completed.second = 0; completed.is_date = completed.is_utc = 1;
-			t.tv_sec = icaltime_as_timet (completed);
-			t.tv_usec = 0;
-			set_named_datetime_value (PidLidTaskDateCompleted, &t);
+			tt = icaltime_as_timet (completed);
+			set_named_datetime_value (PidLidTaskDateCompleted, tt);
 		}
 
 		/* Start */
 		dtstart.hour = dtstart.minute = dtstart.second = 0; dtstart.is_date = dtstart.is_utc = 1;
-		t.tv_sec = icaltime_as_timet (dtstart);
-		t.tv_usec = 0;
+		tt = icaltime_as_timet (dtstart);
 		if (!icaltime_is_null_time (dtstart)) {
-			set_named_datetime_value (PidLidTaskStartDate, &t);
+			set_named_datetime_value (PidLidTaskStartDate, tt);
 		}
 
 		/* Due */
 		dtend.hour = dtend.minute = dtend.second = 0; dtend.is_date = dtend.is_utc = 1;
-		t.tv_sec = icaltime_as_timet (dtend);
-		t.tv_usec = 0;
+		tt = icaltime_as_timet (dtend);
 		if (!icaltime_is_null_time (dtend)) {
-			set_named_datetime_value (PidLidTaskDueDate, &t);
+			set_named_datetime_value (PidLidTaskDueDate, tt);
 		}
 
 		/* FIXME: Evolution does not support recurring tasks */
@@ -2099,3 +2271,28 @@ exchange_mapi_cal_utils_get_props_cb (ExchangeMapiConnection *conn, mapi_id_t fi
 
 	return exchange_mapi_cal_utils_add_named_ids (conn, fid, mem_ctx, props, GPOINTER_TO_INT (data));
 }
+
+gchar *
+exchange_mapi_cal_utils_get_icomp_x_prop (icalcomponent *comp, const gchar *key)
+{
+	icalproperty *xprop;
+
+	/* Find the old one first */
+	xprop = icalcomponent_get_first_property (comp, ICAL_X_PROPERTY);
+
+	while (xprop) {
+		const gchar *str = icalproperty_get_x_name (xprop);
+
+		if (str && !strcmp (str, key)) {
+			break;
+		}
+
+		xprop = icalcomponent_get_next_property (comp, ICAL_X_PROPERTY);
+	}
+
+	if (xprop)
+		return icalproperty_get_value_as_string_r (xprop);
+
+	return NULL;
+}
+
